


import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import javax.persistence.EntityManagerFactory;
import javax.persistence.Persistence;
import javax.transaction.TransactionManager;

import org.drools.KnowledgeBase;
import org.drools.KnowledgeBaseFactory;
import org.drools.SystemEventListenerFactory;
import org.drools.builder.KnowledgeBuilder;
import org.drools.builder.KnowledgeBuilderFactory;
import org.drools.builder.ResourceType;
import org.drools.definition.process.Connection;
import org.drools.definition.process.Node;
import org.drools.definition.process.WorkflowProcess;
import org.drools.event.process.ProcessCompletedEvent;
import org.drools.event.process.ProcessEventListener;
import org.drools.event.process.ProcessNodeLeftEvent;
import org.drools.event.process.ProcessNodeTriggeredEvent;
import org.drools.event.process.ProcessStartedEvent;
import org.drools.event.process.ProcessVariableChangedEvent;
import org.drools.event.rule.DefaultAgendaEventListener;
import org.drools.event.rule.RuleFlowGroupActivatedEvent;
import org.drools.io.ResourceFactory;
import org.drools.logger.KnowledgeRuntimeLogger;
import org.drools.logger.KnowledgeRuntimeLoggerFactory;
import org.drools.persistence.jpa.JPAKnowledgeService;
import org.drools.runtime.Environment;
import org.drools.runtime.EnvironmentName;
import org.drools.runtime.KnowledgeSessionConfiguration;
import org.drools.runtime.StatefulKnowledgeSession;
import org.drools.runtime.process.ProcessInstance;
import org.drools.runtime.rule.FactHandle;
import org.jbpm.process.workitem.wsht.LocalHTWorkItemHandler;
import org.jbpm.task.identity.DefaultUserGroupCallbackImpl;
import org.jbpm.task.identity.UserGroupCallbackManager;
import org.jbpm.task.query.TaskSummary;
import org.jbpm.task.service.TaskServiceSession;
import org.jbpm.task.service.local.LocalTaskService;
import org.jbpm.task.utils.ContentMarshallerHelper;
import org.jbpm.task.utils.OnErrorAction;
import org.jbpm.workflow.core.node.HumanTaskNode;
import org.jbpm.workflow.core.node.Join;
import org.jbpm.workflow.core.node.Split;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import bitronix.tm.Configuration;
import bitronix.tm.TransactionManagerServices;
import bitronix.tm.resource.jdbc.PoolingDataSource;

public class TestFSM2 {
	
	private static Logger logger = LoggerFactory.getLogger(TestFSM.class);
	private static StatefulKnowledgeSession ksession;
	private static PoolingDataSource ds = new PoolingDataSource();
	private static Map<String,List<String>> actionsMap = new HashMap<String,List<String>>();
	
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		
		// configure BTM
	    ds.setUniqueName("jdbc/testDS1");
        ds.setClassName("org.h2.jdbcx.JdbcDataSource");
        ds.setMaxPoolSize(3);
        ds.setAllowLocalTransactions(true);
        ds.getDriverProperties().put("user", "sa");
        ds.getDriverProperties().put("password", "sasa");
        ds.getDriverProperties().put("URL", "jdbc:h2:mem:mydb");
        ds.init();
		Configuration conf = TransactionManagerServices.getConfiguration();
		conf.setServerId("jvm-1");
		conf.setLogPart1Filename("./tx-logs/part1.btm");
		conf.setLogPart2Filename("./tx-logs/part2.btm");
		
		
		try {
			
			
			Environment env = KnowledgeBaseFactory.newEnvironment();
			env.set( EnvironmentName.ENTITY_MANAGER_FACTORY, Persistence.createEntityManagerFactory("org.jbpm.persistence.jpa"));
		
			env.set( EnvironmentName.TRANSACTION_MANAGER,TransactionManagerServices.getTransactionManager());
			Properties properties = new Properties();
			properties.put("drools.processInstanceManagerFactory", "org.jbpm.persistence.processinstance.JPAProcessInstanceManagerFactory");
			properties.put("drools.processSignalManagerFactory", "org.jbpm.persistence.processinstance.JPASignalManagerFactory");
			KnowledgeSessionConfiguration config = KnowledgeBaseFactory.newKnowledgeSessionConfiguration(properties);
			KnowledgeBase kbase = readKnowledgeBase();
			
			//ksession = JPAKnowledgeService.newStatefulKnowledgeSession(kbase, config, env);
			ksession = JPAKnowledgeService.newStatefulKnowledgeSession(kbase, null, env);
			//ksession = kbase.newStatefulKnowledgeSession();

			
	        EntityManagerFactory emf = Persistence.createEntityManagerFactory("org.jbpm.task");
	        org.jbpm.task.service.TaskService taskService = new org.jbpm.task.service.TaskService(emf, SystemEventListenerFactory.getSystemEventListener());
	        TaskServiceSession taskSession = taskService.createSession();
	        
			
			KnowledgeRuntimeLogger logger = KnowledgeRuntimeLoggerFactory.newThreadedFileLogger(ksession, "test", 1000);
			CustomProcessEventListener customProcessEventListener = new CustomProcessEventListener();
			ksession.addEventListener(customProcessEventListener);
			//**********************//
			ksession.addEventListener(new DefaultAgendaEventListener(){

				@Override
				public void afterRuleFlowGroupActivated(
						RuleFlowGroupActivatedEvent event) {
					 ((StatefulKnowledgeSession) event.getKnowledgeRuntime())
                     .fireAllRules();
				}
				
			});
			
			System.out.println("Registering Human Task..");
			
			//**********************//
			LocalTaskService localTaskService = new LocalTaskService(taskService);
			LocalHTWorkItemHandler localHTWorkItemHandler = new LocalHTWorkItemHandler(localTaskService,ksession, OnErrorAction.RETHROW);
			//SyncWSHumanTaskHandler  localHTWorkItemHandler = new SyncWSHumanTaskHandler(localTaskService,ksession);
			//GenericHTWorkItemHandler localHTWorkItemHandler = new GenericHTWorkItemHandler(ksession);
			localHTWorkItemHandler.setClient(localTaskService);
			//localHTWorkItemHandler.setIpAddress("127.0.0.1");
			//localHTWorkItemHandler.setPort(9123);
			localHTWorkItemHandler.setLocal(true);
			localHTWorkItemHandler.connect();
			ksession.getWorkItemManager().registerWorkItemHandler("Human Task", localHTWorkItemHandler);
			
			
			Properties userGroups = new Properties();
			userGroups.setProperty("john", "user");
			UserGroupCallbackManager manager = UserGroupCallbackManager.getInstance();
			manager.setCallback(new DefaultUserGroupCallbackImpl(userGroups));

			TransactionManager tm = TransactionManagerServices.getTransactionManager();
			// start a new process instance
			Map<String, Object> params = new HashMap<String, Object>();
			String userId = "john";
			params.put("userId", userId);
			tm.begin();
			ProcessInstance processInstance = ksession.startProcess("HumanTask", params);
			ksession.insert(processInstance);
			ksession.fireAllRules();
			tm.commit();
//			retrievePossibleActions(kbase,processInstance.getProcessId());
			
			
			List<TaskSummary> taskSummaryList = localTaskService.getTasksAssignedAsPotentialOwner(userId, "en-UK");
			boolean terminated = false;
			do{
				for (TaskSummary taskSummary : taskSummaryList) {
					//bypass historic tasks
					if(taskSummary.getProcessInstanceId()!= processInstance.getId())
						continue;
					
					long taskId = taskSummary.getId();
					System.out.println(taskSummary.getId() + "\t" + taskSummary.getName() + "\t" + taskSummary.getStatus());
					
					List<String> actions = actionsMap.get(taskSummary.getName());
//					if(actions.size()>0){
//						System.out.println("You can take one of the following actions:");
//						int i=1;
//						for(String action : actions){
//							System.out.println(i++ +")"+action);
//						}
//						System.out.println("Please enter your chooice:");
//						int chooice = 0;
//						do{
//							chooice = getInt();
//						}while(chooice<=0 && chooice >=actions.size());
						
//						localTaskService.claim(taskId, userId);
						localTaskService.start(taskId, userId);
						Map<String,Object> data = new HashMap<String,Object>();
						data.put("action", "");
						localTaskService.completeWithResults(taskSummary.getId(), userId,data );	
						taskSummaryList = localTaskService.getTasksAssignedAsPotentialOwner(userId, "en-UK");
//					}
//					else{
//						System.out.println("There is no action for current status.");
//						terminated = true;
//					}
					
				}
				
				taskSummaryList = localTaskService.getTasksAssignedAsPotentialOwner(userId, "en-UK");
			}while(taskSummaryList.size()>0 && terminated==false);	
		
			System.out.println("Process is " + ksession.getProcessInstance(processInstance.getId()));
			
			logger.close();
			System.out.println((new Date()) + "\tProcess completed?");
			System.exit(0);
		} catch (Throwable t) {
			t.printStackTrace();
		}
	}
	
	public static int getInt() throws java.io.IOException {
		BufferedReader in = new BufferedReader(new InputStreamReader(System.in));
		while (true) {
			try {
				String s = (in.readLine()).trim();
				int y = (new Integer(s)).intValue();
				return y;
			} catch (java.lang.NumberFormatException e) {
			}
		}
	}
	
	/**
	 * retrieve all possible actions
	 * @param kbase
	 * @param processId
	 */
	private static void retrievePossibleActions(KnowledgeBase kbase, String processId){
		//TODO handle all kinds of situations.  
		WorkflowProcess workFlowProcess = ((WorkflowProcess) kbase.getProcess(processId));
        for (org.drools.definition.process.Node node :  workFlowProcess.getNodes()){
        	if(node instanceof HumanTaskNode){
            	List<String> actions = new ArrayList<String>();
            	//TODO use taskName, instead of name
            	actionsMap.put(node.getName(), actions);
        		for(String outgoing   : node.getOutgoingConnections().keySet()){
        			for(Connection connection   :node.getOutgoingConnections().get(outgoing)){
        				Node toNode = connection.getTo();
        				for(String outgoing1   : toNode.getOutgoingConnections().keySet()){
        					for(Connection connection1  : toNode.getOutgoingConnections().get(outgoing1)){
        						Node toNode1 = connection1.getTo();
                				if(toNode1 instanceof HumanTaskNode){
                					actions.add(((Split)connection1.getFrom()).getConstraint(connection1).getName());
                				}
                				else if(toNode1 instanceof Join){
                					for(Connection connection2  : toNode1.getOutgoingConnections().get(outgoing1)){
                						Node toNode2 = connection2.getTo();
                        				if(toNode2 instanceof HumanTaskNode){
                        					actions.add(((Split)connection1.getFrom()).getConstraint(connection1).getName());
                        				}
                					}
                				}
        					}
        				}
        				
        			}
        		}
        	}
        }
	}
	
	private static KnowledgeBase readKnowledgeBase() throws Exception {
		System.out.println("Loading knowledge base.");
		KnowledgeBuilder kbuilder = KnowledgeBuilderFactory.newKnowledgeBuilder();
//		kbuilder.add(ResourceFactory.newClassPathResource("ApplicationStatus.bpmn"), ResourceType.BPMN2);
		kbuilder.add(ResourceFactory.newClassPathResource("HumanTask.bpmn"), ResourceType.BPMN2);
		
		return kbuilder.newKnowledgeBase();
	}
	
	private static class CustomProcessEventListener implements ProcessEventListener {


		public void afterNodeLeft(ProcessNodeLeftEvent arg0) {
			// TODO Auto-generated method stub
			
		}

		public void afterNodeTriggered(ProcessNodeTriggeredEvent arg0) {
			// TODO Auto-generated method stub
			
		}

		public void afterProcessCompleted(ProcessCompletedEvent event) {
			// TODO Auto-generated method stub
			//logger.debug("Process has completed.");
			//ksession.dispose();
			//System.exit(0);
		}

		public void afterProcessStarted(ProcessStartedEvent arg0) {
			// TODO Auto-generated method stub
			
		}

		public void afterVariableChanged(ProcessVariableChangedEvent event) {
	        FactHandle handle = event.getKnowledgeRuntime().getFactHandle( event.getProcessInstance() );
		        if(handle != null){
	             event.getKnowledgeRuntime().update( handle, event.getProcessInstance() );
	        }
	        else{
	        	event.getKnowledgeRuntime().insert(event.getProcessInstance());
	        }
		}

		public void beforeNodeLeft(ProcessNodeLeftEvent arg0) {
			// TODO Auto-generated method stub
			
		}

		public void beforeNodeTriggered(ProcessNodeTriggeredEvent arg0) {
			// TODO Auto-generated method stub
			
		}

		public void beforeProcessCompleted(ProcessCompletedEvent arg0) {
			// TODO Auto-generated method stub
			logger.debug("Process is going to be completed.");
		}

		public void beforeProcessStarted(ProcessStartedEvent arg0) {
			// TODO Auto-generated method stub
			
		}

		public void beforeVariableChanged(ProcessVariableChangedEvent arg0) {
			// TODO Auto-generated method stub
			
		}
		
	}
	
	private static class SystemEventListener implements org.drools.SystemEventListener {
		public void debug(String arg0) {
			logger.debug(arg0);
		}
		public void debug(String arg0, Object arg1) {
			logger.debug(arg0);
		}
		public void exception(Throwable arg0) {
			logger.error("",arg0);
		}
		public void exception(String arg0, Throwable arg1) {
			logger.error(arg0, arg1);
		}
		public void info(String arg0) {
			logger.info(arg0);
		}
		public void info(String arg0, Object arg1) {
			logger.info(arg0);
		}
		public void warning(String arg0) {
			logger.warn(arg0);
		}
		public void warning(String arg0, Object arg1) {
			logger.warn(arg0);
		}
	}
	
	//**********************//
	public  static Map<String,Object> getTaskParameters(LocalTaskService localTaskService ,long taskId) {
		org.jbpm.task.Task t = localTaskService.getTask(taskId);

		Object data = ContentMarshallerHelper.unmarshall(localTaskService
				.getContent(t.getTaskData().getDocumentContentId())
				.getContent(), ksession.getEnvironment());

		System.out.println("parameter = " + data);

		return (Map<String,Object>)data;
	}

}
